import numpy as np
from scipy.optimize import minimize


def expected_ucb(res, n_random_starts=100, alpha=1.96, random_state=None):
    """ Compute the expected (pessimistic) optimum of the optimization result.

    This will compute `gp_mean + alpha * gp_se` and find the optimum of that function.

    Parameters
    ----------
    * `res`  [`OptimizeResult`, scipy object]:
        The optimization result returned by a `bask` minimizer.
    * `n_random_starts` [int, default=100]:
        The number of random starts for the minimization of the surrogate model.
    * `alpha` [float, default=1.96]:
        Number of standard errors to add to each point.
    * `random_state` [int, RandomState instance, or None (default)]:
        Set random state to something other than None for reproducible
        results.
    Returns
    -------
    * `x` [list]: location of the minimum.
    * `fun` [float]: the surrogate function value at the minimum.
    """

    def func(x):
        reg = res.models[-1]
        mu, std = reg.predict(x.reshape(1, -1), return_std=True)
        return (mu + alpha * std)[0]

    xs = [res.space.transform([res.x]).tolist()]
    if n_random_starts > 0:
        xs.extend(
            res.space.transform(
                res.space.rvs(n_random_starts, random_state=random_state)
            ).tolist()
        )

    best_x = None
    best_fun = np.inf

    for x0 in xs:
        r = minimize(func, x0=x0, bounds=[(0.0, 1.0)] * len(res.space.bounds))

        if r.fun < best_fun:
            best_x = r.x
            best_fun = r.fun
    return res.space.inverse_transform(best_x[None, :])[0], best_fun



"""The server worker, which reads results and schedules new jobs."""
import json
import logging
import os
import re
import sys
from datetime import datetime
from time import sleep

import joblib
import numpy as np
import pandas as pd
import scipy.stats
from bask import Optimizer
from skopt.learning.gaussian_process.kernels import Matern, ConstantKernel
from skopt.space import space as skspace
from skopt.utils import normalize_dimensions, create_result
from sqlalchemy.orm import sessionmaker

from tune.priors import roundflat
from tune.utils import expected_ucb
from tune.db_workers.dbmodels import (
    Base,
    SqlJob,
    SqlUCIParam,
    SqlResult,
    SqlTimeControl,
    SqlTune,
)
from tune.db_workers.utils import (
    get_session_maker,
    create_sqlalchemy_engine,
    TimeControl,
    simple_penta_to_score,
)
from tune.io import InitStrings

__all__ = ["TuningServer"]


class TuningServer(object):
    def __init__(self, experiment_path, dbconfig_path, **kwargs):
        self.logger = logging.getLogger("TuningServer")
        self.experiment_path = experiment_path
        if os.path.isfile(dbconfig_path):
            with open(dbconfig_path, "r") as config_file:
                config = config_file.read().replace("\n", "")
                self.logger.debug(f"Reading DB config:\n{config}")
                self.connect_params = json.loads(config)
        else:
            raise ValueError("No dbconfig file found at provided path")

        self.engine = create_sqlalchemy_engine(self.connect_params)
        Base.metadata.create_all(self.engine)
        sm = sessionmaker(bind=self.engine)
        self.sessionmaker = get_session_maker(sm)

        if os.path.isfile(experiment_path):
            with open(experiment_path, "r+") as experiment_file:
                exp = experiment_file.read().replace("\n", "")
                self.logger.debug(f"Reading experiment config:\n{exp}")
                self.experiment = json.loads(exp)
                self.logger.debug(f"self.experiment = \n{self.experiment}")
        else:
            raise ValueError("No experiment config file found at provided path")
        self.time_controls = [
            TimeControl.from_strings(*x) for x in self.experiment["time_controls"]
        ]
        self.rng = np.random.RandomState(self.experiment.get("random_seed", 123))
        self.setup_tuner()

        try:
            os.makedirs("experiments")
        except FileExistsError:
            pass
        # TODO: in principle after deleting all jobs from the database,
        #       this could be problematic:
        self.pos = None
        self.chain = None
        if "tune_id" in self.experiment:
            self.resume_tuning()

    def write_experiment_file(self):
        with open(self.experiment_path, "w") as experiment_file:
            experiment_file.write(json.dumps(self.experiment, indent=2))

    def save_state(self):
        path = os.path.join(
            "experiments", f"data_tuneid_{self.experiment['tune_id']}.npz"
        )
        np.savez_compressed(
            path, np.array(self.opt.gp.pos_), np.array(self.opt.gp.chain_)
        )

    def resume_tuning(self):
        path = os.path.join(
            "experiments", f"data_tuneid_{self.experiment['tune_id']}.npz"
        )
        if os.path.exists(path):
            data = np.load(path)
            self.opt.gp.pos_ = data["arr_0"]
            self.opt.gp.chain_ = data["arr_1"]

    def parse_dimensions(self, param_dict):
        def make_numeric(s):
            try:
                return int(s)
            except ValueError:
                try:
                    return float(s)
                except ValueError:
                    return s

        dimensions = []
        for s in param_dict.values():
            prior_str = re.findall(r"(\w+)\(", s)[0]
            prior_param_strings = re.findall(r"\((.*?)\)", s)[0].split(",")
            keys = [x.split("=")[0].strip() for x in prior_param_strings]
            vals = [make_numeric(x.split("=")[1].strip()) for x in prior_param_strings]
            dim = getattr(skspace, prior_str)(**dict(zip(keys, vals)))
            dimensions.append(dim)
        return dimensions

    def parse_priors(self, priors):
        if isinstance(priors, str):
            try:
                result = joblib.load(priors)
            except IOError:
                self.logger.error(
                    f"Priors could not be loaded from path {priors}. Terminating..."
                )
                sys.exit(1)
        else:
            result = []
            for i, p in enumerate(priors):
                prior_str = re.findall(r"(\w+)\(", p)[0]
                prior_param_strings = re.findall(r"\((.*?)\)", p)[0].split(",")
                keys = [x.split("=")[0].strip() for x in prior_param_strings]
                vals = [float(x.split("=")[1].strip()) for x in prior_param_strings]
                if prior_str == "roundflat":
                    prior = lambda x, keys=keys, vals=vals: roundflat(np.exp(x), **dict(zip(keys, vals))) + x
                else:
                    dist = getattr(scipy.stats, prior_str)(**dict(zip(keys, vals)))
                    if i == 0 or i == len(priors) - 1:
                        # The signal variance and the signal noise are in positive, sqrt domain
                        prior = (
                            lambda x, dist=dist: dist.logpdf(np.sqrt(np.exp(x)))
                            + x / 2.0
                            - np.log(2.0)
                        )
                    else:
                        # The lengthscale(s) are in positive domain
                        prior = lambda x, dist=dist: dist.logpdf(np.exp(x)) + x
                result.append(prior)
        return result

    def setup_tuner(self):
        self.tunecfg = self.experiment["tuner"]
        self.parameters = list(self.tunecfg["parameters"].keys())
        self.dimensions = self.parse_dimensions(self.tunecfg["parameters"])
        self.space = normalize_dimensions(self.dimensions)
        self.priors = self.parse_priors(self.tunecfg["priors"])

        self.kernel = ConstantKernel(
            constant_value=self.tunecfg.get("variance_value", 0.1 ** 2),
            constant_value_bounds=tuple(
                self.tunecfg.get("variance_bounds", (0.01 ** 2, 0.5 ** 2))
            ),
        ) * Matern(
            length_scale=self.tunecfg.get("length_scale_value", 0.3),
            length_scale_bounds=tuple(
                self.tunecfg.get("length_scale_bounds", (0.2, 0.8))
            ),
            nu=2.5,
        )
        self.opt = Optimizer(
            dimensions=self.dimensions,
            n_points=self.tunecfg.get("n_points", 1000),
            n_initial_points=self.tunecfg.get(
                "n_initial_points", 5 * len(self.dimensions)
            ),
            gp_kernel=self.kernel,
            gp_kwargs=dict(normalize_y=True),
            gp_priors=self.priors,
            acq_func=self.tunecfg.get("acq_func", "ts"),
            acq_func_kwargs=self.tunecfg.get(
                "acq_func_kwargs", None
            ),  # TODO: Check if this works for all parameters
            random_state=self.rng.randint(0, np.iinfo(np.int32).max),
        )

    def query_data(self, session, include_active=False):
        tune_id = self.experiment["tune_id"]
        # First check if samplesize was reached:
        sample_sizes = np.array(
            session.query(
                SqlResult.ww_count
                + SqlResult.wd_count
                + SqlResult.wl_count
                + SqlResult.dd_count
                + SqlResult.dl_count
                + SqlResult.ll_count
            )
            .join(SqlJob)
            .filter(SqlJob.active, SqlJob.tune_id == tune_id)
            .all()
        ).squeeze()
        samplesize_reached = False
        if np.all(sample_sizes >= self.experiment.get("minimum_samplesize", 16)):
            samplesize_reached = True

        q = session.query(SqlJob).filter(SqlJob.tune_id == tune_id)
        if not include_active:
            q = q.filter(SqlJob.active == False)  # noqa
        jobs = q.all()
        query = session.query(SqlUCIParam.job_id, SqlUCIParam.key, SqlUCIParam.value).join(SqlJob).filter(SqlJob.tune_id == tune_id)
        df = pd.read_sql(query.statement, query.session.bind)
        df['value'] = df['value'].astype(float)
        self.logger.debug(f"Data frame: {df.head()}")
        X = df.pivot(index="job_id", columns="key").sort_index().droplevel(0, axis=1)[self.parameters].values
        y = {tc: [] for tc in self.time_controls}
        for job in jobs:
            for result in job.results:
                tc = result.time_control.to_tuple()
                if tc not in self.time_controls:
                    continue
                draw_rate = float(result.time_control.draw_rate)
                counts = np.array(
                    [
                        result.ll_count,
                        result.dl_count,
                        result.wl_count + result.dd_count,
                        result.wd_count,
                        result.ww_count,
                    ]
                )
                score = simple_penta_to_score(
                    draw_rate=draw_rate, counts=counts,
                    prior_games=self.experiment.get("prior_games", 1),
                    prior_elo=self.experiment.get("prior_elo", 0.0)
                )
                y[tc].append(-score)
        return X, np.array(list(y.values())).mean(axis=0), samplesize_reached

    @staticmethod
    def change_engine_config(engine_config, params):
        init_strings = InitStrings(
            engine_config[0]["initStrings"]
        )  # TODO: allow tuning of different index
        for k, v in params.items():
            init_strings[k] = v

    def insert_jobs(self, session, new_x):
        # First set all active jobs to inactive:
        session.query(SqlJob).filter(
            SqlJob.active == True, SqlJob.tune_id == self.experiment["tune_id"]
        ).update(
            {"active": False}
        )  # noqa

        # Insert new job:
        job_dict = {
            "engine": self.experiment["engine"],
            "cutechess": self.experiment["cutechess"],
        }
        job_json = json.dumps(job_dict)
        job = SqlJob(
            minimum_version=self.experiment.get("minimum_version", 1),
            maximum_version=self.experiment.get("maximum_version", None),
            minimum_samplesize=self.experiment.get("minimum_samplesize", 16),
            config=job_json,
            engine1_exe=self.experiment.get("engine1_exe", "lc0"),
            engine1_nps=self.experiment["engine1_nps"],
            engine2_exe=self.experiment.get("engine2_exe", "sf"),
            engine2_nps=self.experiment["engine2_nps"],
            tune_id=self.experiment["tune_id"],
        )
        session.add(job)
        for i, tc in enumerate(self.time_controls):
            sql_tc = (
                session.query(SqlTimeControl)
                .filter(
                    SqlTimeControl.engine1_time == tc.engine1_time,
                    SqlTimeControl.engine1_increment == tc.engine1_increment,
                    SqlTimeControl.engine2_time == tc.engine2_time,
                    SqlTimeControl.engine2_increment == tc.engine2_increment,
                )
                .one_or_none()
            )
            if sql_tc is None:
                sql_tc = SqlTimeControl(
                    engine1_time=tc.engine1_time,
                    engine1_increment=tc.engine1_increment,
                    engine2_time=tc.engine2_time,
                    engine2_increment=tc.engine2_increment,
                )
                session.add(sql_tc)

            result = SqlResult(job=job, time_control=sql_tc)
            session.add(result)
        for k, v in zip(self.parameters, new_x):
            param = SqlUCIParam(key=k, value=str(v), job=job)
            session.add(param)

    def run(self):
        # 0. Before we run the main loop, do we need to initialize or resume?
        #    * Resume from files (in experiment folder)
        #    * Create tune entry in db if it does not exist yet

        if "tune_id" not in self.experiment:
            with self.sessionmaker() as session:
                tune = SqlTune(
                    weight=self.experiment.get("weight", 1.0),
                    description=self.experiment.get("description", None),
                )
                session.add(tune)
                session.flush()
                self.experiment["tune_id"] = tune.id
                self.write_experiment_file()
                new_x = self.opt.ask()
                # Alter engine json using Initstrings
                params = dict(zip(self.parameters, new_x))
                self.change_engine_config(self.experiment["engine"], params)
                self.insert_jobs(session, new_x)
                self.logger.info("New jobs committed to database.")
        while True:
            self.logger.debug("Begin querying for new data...")
            # Check if minimum sample size and minimum wait time are reached, then query data and update model:
            with self.sessionmaker() as session:
                X, y, samplesize_reached = self.query_data(session, include_active=True)
                self.logger.debug(
                    f"Queried the database for data and got (last 5):\n{X[-5:]}\n{y[-5:]}"
                )
                if len(X) == 0:
                    self.logger.info("There are no datapoints yet, start first job")
                    new_x = self.opt.ask()
                    # Alter engine json using Initstrings
                    params = dict(zip(self.parameters, new_x))
                    self.change_engine_config(self.experiment["engine"], params)
                    self.insert_jobs(session, new_x)
                    self.logger.info("New jobs committed to database.")
                    samplesize_reached = False

            if not samplesize_reached:
                sleep_seconds = self.experiment.get("sleep_time", 60)
                self.logger.debug(
                    f"Required sample size not yet reached. Sleeping {sleep_seconds} seconds."
                )
                sleep(sleep_seconds)
                continue

            # Tell optimizer about the new results:
            now = datetime.now()
            self.opt.tell(
                X.tolist(),
                y.tolist(),
                fit=True,
                replace=True,
                n_samples=self.tunecfg["n_samples"],
                gp_samples=self.tunecfg["gp_samples"],
                gp_burnin=self.tunecfg["gp_burnin"],
                progress=False,
            )
            later = datetime.now()
            difference = (later - now).total_seconds()
            self.logger.info(
                f"Calculating GP posterior and acquisition function finished in {difference}s"
            )
            self.logger.info(f"Current GP kernel:\n{self.opt.gp.kernel_}")
            if self.opt.gp.chain_ is not None:
                self.logger.debug("Saving position and chain")
                self.save_state()

            # Ask optimizer for new configuration and insert jobs:
            new_x = self.opt.ask()
            # Alter engine json using Initstrings
            params = dict(zip(self.parameters, new_x))
            self.change_engine_config(self.experiment["engine"], params)
            with self.sessionmaker() as session:
                self.insert_jobs(session, new_x)
            self.logger.info("New jobs committed to database.")
            sleep(self.experiment.get("sleep_time", 60))

            result_object = create_result(
                Xi=X.tolist(), yi=y.tolist(), space=self.opt.space, models=[self.opt.gp]
            )
            try:
                opt_x, opt_y = expected_ucb(result_object)
                self.logger.info(
                    f"Current optimum: {dict(zip(self.parameters, np.around(opt_x,4)))}"
                )
            except ValueError:
                self.logger.info(f"Current optimum: None (optimizer errored out :( )")

    def deactivate(self):
        raise NotImplementedError

    def reactivate(self):
        raise NotImplementedError



"""Console script for chess_tuning_tools."""
import sys
import click
import logging

from .db_workers import TuningServer, TuningClient


@click.group()
def cli():
    pass


@cli.command()
@click.option("--verbose", "-v", is_flag=True, default=False, help="Turn on debug output.")
@click.option("--logfile", default=None, help="Path to where the log is saved to.")
@click.option("--terminate-after", default=0, help="Terminate the client after x minutes.")
@click.option("--clientconfig", default=None, help="Path to the client configuration file.")
@click.argument("dbconfig")
def run_client(verbose, logfile, terminate_after, clientconfig, dbconfig):
    """ Run the client to generate games for distributed tuning.

    In order to connect to the database you need to provide a valid DBCONFIG
    json file. It contains the necessary parameters to connect to the database
    where it can fetch jobs and store results.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level, filename=logfile, format="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    tc = TuningClient(dbconfig_path=dbconfig, terminate_after=terminate_after, clientconfig=clientconfig)
    tc.run()


@cli.command()
@click.option("--verbose", "-v", is_flag=True, default=False, help="Turn on debug output.")
@click.option("--logfile", default=None, help="Path to where the log is saved to.")
@click.argument("command")
@click.argument("experiment_file")
@click.argument("dbconfig")
def run_server(verbose, logfile, command, experiment_file, dbconfig):
    """Run the tuning server for a given EXPERIMENT_FILE (json).

    To connect to the database you also need to provide a DBCONFIG json file.

    \b
    You can choose from these COMMANDs:
     * run: Starts the server.
     * deactivate: Deactivates all active jobs of the given experiment.
     * reactivate: Reactivates all recent jobs for which sample size is not reached yet.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=log_level, filename=logfile, format="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    tc = TuningServer(experiment_path=experiment_file, dbconfig_path=dbconfig)
    if command == "run":
        tc.run()
    elif command == "deactivate":
        tc.deactivate()
    elif command == "reactivate":
        tc.reactivate()
    else:
        raise ValueError(f"Command {command} is not recognized. Terminating...")


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import numpy as np
from numpy.testing import assert_almost_equal
from tune.priors import roundflat


def test_roundflat():
    assert_almost_equal(roundflat(0.3), 0.0, decimal=0.1)

    assert roundflat(0.0) == -np.inf
    assert roundflat(-1.0) == -np.inf



#!/usr/bin/env python

from decimal import Decimal

import numpy as np
from numpy.testing import assert_almost_equal

from tune.db_workers.utils import (
    penta,
    ldw_probabilities,
    draw_rate_to_elo,
    compute_probabilities_for_bias,
    compute_probabilities,
    elo_to_bayeselo,
    penta_to_score,
    TimeControl,
)


def test_penta():
    ldw1 = np.array([0.1, 0.2, 0.7])
    ldw2 = np.array([0.2, 0.2, 0.6])

    result = penta(ldw1, ldw2)
    expected = np.array([0.02, 0.06, 0.24, 0.26, 0.42])
    assert_almost_equal(result, expected, decimal=3)


def test_ldw_probabilities():
    result = ldw_probabilities(elo=50, draw_elo=200, bias=200)
    expected = np.array([0.06975828735890623, 0.35877859523271227, 0.5714631174083815])
    assert_almost_equal(result, expected)


def test_draw_rate_to_elo():
    result = draw_rate_to_elo(0.5)
    expected = np.array(190.84850188786498)
    assert_almost_equal(result, expected)


def test_compute_probabilities_for_bias():
    result = compute_probabilities_for_bias(elo=50, draw_elo=200, bias=200)
    expected = np.array([0.029894, 0.18540627, 0.41591514, 0.30154527, 0.06723932])
    assert_almost_equal(result, expected)


def test_compute_probabilities():
    result = compute_probabilities(elo=50, draw_elo=200, biases=(0, 200))
    expected = np.array(
        [
            0.033318056828286285,
            0.19078749500997028,
            0.3957332350793835,
            0.30255132321508954,
            0.07760988986727044,
        ]
    )
    assert_almost_equal(result, expected)


def test_elo_to_bayeselo():
    result = elo_to_bayeselo(elo=50, draw_elo=200, biases=(0, 200))
    expected = 71.513929
    assert_almost_equal(result, expected, decimal=5)


def test_penta_to_score():
    counts = np.array([1, 2, 3, 4, 5])
    result = penta_to_score(draw_rate=0.5, counts=counts, prior_games=10, prior_elo=0)
    expected = 0.4016368226279837
    assert_almost_equal(result, expected)


def test_timecontrol():
    strings = ("3.0+0.03", "7.0+0.0")
    result = TimeControl.from_strings(*strings)
    expected = (Decimal("3.0"), Decimal("0.03"), Decimal(7), Decimal(0))
    assert result == expected

    assert result.to_strings() == strings



import re
import sys
from collections.abc import MutableMapping


# TODO: Backup file to restore it, should there be an error
def uci_tuple(uci_string):
    try:
        name, value = re.findall(r"name (\w+) value (\S*)", uci_string)[0]
    except IndexError:
        print(f"Error parsing UCI tuples:\n{uci_string}")
        sys.exit(1)
    try:
        tmp = float(value)
    except ValueError:
        tmp = value
    return name, tmp


def set_option(name, value):
    return f"setoption name {name} value {value}"


class InitStrings(MutableMapping):
    def __init__(self, init_strings):
        self._init_strings = init_strings

    def __len__(self):
        return len(self._init_strings)

    def __getitem__(self, key):
        for s in self._init_strings:
            name, value = uci_tuple(s)
            if key == name:
                return value
        raise KeyError(key)

    def __setitem__(self, key, value):
        for i, s in enumerate(self._init_strings):
            name, _ = uci_tuple(s)
            if key == name:
                self._init_strings[i] = set_option(key, value)
                return
        self._init_strings.append(set_option(key, value))

    def __delitem__(self, key):
        elem = -1
        for i, s in enumerate(self._init_strings):
            name, _ = uci_tuple(s)
            if key == name:
                elem = i
                break
        if elem != -1:
            del self._init_strings[i]
        else:
            raise KeyError(key)

    def __contains__(self, key):
        for s in self._init_strings:
            name, _ = uci_tuple(s)
            if key == name:
                return True
        return False

    def __iter__(self):
        for s in self._init_strings:
            name, _ = uci_tuple(s)
            yield name

    def __repr__(self):
        return repr(self._init_strings)



from datetime import datetime

from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    JSON,
    Numeric,
    PrimaryKeyConstraint,
    String,
    Table,
)
from sqlalchemy.orm import relationship

from tune.db_workers.dbmodels import Base
from tune.db_workers.utils import TimeControl

__all__ = [
    "SqlTune",
    "SqlJob",
    "SqlUCIParam",
    "SqlTimeControl",
    "SqlResult",
]

SCHEMA = "new"


class SqlTune(Base):
    __tablename__ = "tunes"
    __table_args__ = {"schema": SCHEMA}
    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, default=datetime.utcnow, nullable=False)
    weight = Column(Numeric, default=1.0, nullable=False)
    description = Column(String(250), nullable=True)

    jobs = relationship("SqlJob", back_populates="tune", cascade="all")

    def __repr__(self):
        return (
            f"<Tune (id={self.id}, timestamp={self.timestamp}, weight={self.weight})>"
        )


job_tc_table = Table(
    "jobstotimes",
    Base.metadata,
    Column("job_id", Integer, ForeignKey(f"{SCHEMA}.jobs.id")),
    Column("tc_id", Integer, ForeignKey(f"{SCHEMA}.timecontrols.id")),
    schema=SCHEMA
)


class SqlJob(Base):
    __tablename__ = "jobs"
    __table_args__ = {"schema": SCHEMA}
    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, default=datetime.utcnow, nullable=False)
    active = Column(Boolean, default=True, nullable=False)
    weight = Column(Numeric, default=1.0, nullable=False)
    minimum_version = Column(Integer, default=1, nullable=False)
    maximum_version = Column(Integer, nullable=True, default=None)
    minimum_samplesize = Column(Integer, nullable=False, default=16)
    config = Column(JSON)
    engine1_exe = Column(String(100), nullable=False, default="lc0")
    engine1_nps = Column(Numeric, nullable=False)
    engine2_exe = Column(String(100), nullable=False, default="sf")
    engine2_nps = Column(Numeric, nullable=False)

    tune_id = Column(Integer, ForeignKey(f"{SCHEMA}.tunes.id"))
    tune = relationship("SqlTune", back_populates="jobs")
    params = relationship("SqlUCIParam", back_populates="job", cascade="all")
    time_controls = relationship("SqlTimeControl", secondary=job_tc_table)
    results = relationship("SqlResult", back_populates="job")

    def __repr__(self):
        return (
            f"<Job (id={self.id}, timestamp={self.timestamp}, active={self.active},"
            f" weight={self.weight}, engine1_exe={self.engine1_exe}, engine1_nps={self.engine1_nps},"
            f" engine2_exe={self.engine2_exe}, engine2_nps={self.engine2_nps}, config={self.config})>"
        )


class SqlUCIParam(Base):
    __tablename__ = "params"
    __table_args__ = (
        PrimaryKeyConstraint("key", "job_id", name="param_pk"),
        {"schema": SCHEMA},
    )
    key = Column(String(100))
    value = Column(String(250), nullable=False)
    job_id = Column(Integer, ForeignKey(f"{SCHEMA}.jobs.id"))
    job = relationship("SqlJob", back_populates="params")

    def __repr__(self):
        return f"<UCI (key={self.key}, value={self.value}, job_id={self.job_id})>"


class SqlTimeControl(Base):
    __tablename__ = "timecontrols"
    __table_args__ = ({"schema": SCHEMA},)
    id = Column(Integer, primary_key=True)
    engine1_time = Column(Numeric, nullable=False)
    engine1_increment = Column(Numeric, nullable=True, default=None)
    engine2_time = Column(Numeric, nullable=False)
    engine2_increment = Column(Numeric, nullable=True, default=None)
    draw_rate = Column(Numeric, nullable=False, default=0.33)

    def __repr__(self):
        engine1_inc = (
            "" if self.engine1_increment is None else f"+{self.engine1_increment}"
        )
        engine2_inc = (
            "" if self.engine2_increment is None else f"+{self.engine2_increment}"
        )
        return (
            f"<TC (id={self.id}, engine1={self.engine1_time}{engine1_inc},"
            f" engine2={self.engine2_time}{engine2_inc}, draw_rate={self.draw_rate})>"
        )

    def to_tuple(self):
        return TimeControl(
            engine1_time=self.engine1_time,
            engine1_increment=self.engine1_increment,
            engine2_time=self.engine2_time,
            engine2_increment=self.engine2_increment
        )


class SqlResult(Base):
    __tablename__ = "results"
    __table_args__ = ({"schema": SCHEMA},)
    job_id = Column(Integer, ForeignKey(f"{SCHEMA}.jobs.id"), primary_key=True)
    tc_id = Column(Integer, ForeignKey(f"{SCHEMA}.timecontrols.id"), primary_key=True)
    ww_count = Column(Integer, default=0)
    wl_count = Column(Integer, default=0)
    wd_count = Column(Integer, default=0)
    dd_count = Column(Integer, default=0)
    dl_count = Column(Integer, default=0)
    ll_count = Column(Integer, default=0)
    job = relationship("SqlJob", back_populates="results")
    time_control = relationship("SqlTimeControl")

    def __repr__(self):
        return (
            f"<Result (job_id={self.job_id}, tc_id={self.tc_id},"
            f" ww={self.ww_count}, wd={self.wd_count},"
            f" wl={self.wl_count}, dd={self.dd_count},"
            f" dl={self.dl_count}, ll={self.ll_count})>"
        )



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "numpy>=1.18",
    "scipy>=1.3.2",
    "pytz",
    "joblib",
    "scikit-optimize",
    "emcee>=3.0.2",
    "psycopg2",
    "bask>=0.1.0",
    "sqlalchemy>=1.3",
    "pandas>=1.0.1"
]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest>=3"]

setup(
    author="Karlson Pfannschmidt",
    author_email="kiudee@mail.upb.de",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A collection of tools for local and distributed tuning of chess engines.",
    entry_points={"console_scripts": ["tune=tune.cli:cli"]},
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="chess, tuning, optimization, engine",
    name="chess-tuning-tools",
    packages=find_packages(include=["tune", "tune.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/kiudee/chess-tuning-tools",
    version="0.3.0",
    zip_safe=False,
)
